import logging

from typing import Union, List, Optional

from ldap.ldapobject import SimpleLDAPObject as LdapObject

from . import _ldaps
from ._object_classes import OrganizationalUnit, create_records
from ._query import object_exists, get, modify, move, ldap_escape, _encode_attributes

logger = logging.getLogger(__name__)
class_plural = "organizationalUnits"
object_class = "organizationalUnit"


@create_records(OrganizationalUnit)
def _get(
    connection: LdapObject, search_base: Optional[str] = None, search_scope: Optional[int] = None, **query_params
) -> List[OrganizationalUnit]:
    return get(
        "(objectClass=organizationalUnit)",
        connection,
        search_base=search_base,
        search_scope=search_scope,
        **query_params,
    )


def _create(connection: LdapObject, distinguished_name: str, name: str, **extra_params) -> None:
    attr_dict: dict = {
        "distinguishedName": bytes(ldap_escape(distinguished_name, is_distinguished_name=True), "utf-8"),
        "name": bytes(ldap_escape(name), "utf-8"),
        "objectClass": [b"top", bytes(ldap_escape(object_class), "utf-8")],
        "ou": bytes(ldap_escape(name), "utf-8"),
    }
    attr_dict.update(_encode_attributes(extra_params))
    return _ldaps._create(connection, ldap_escape(distinguished_name, is_distinguished_name=True), attr_dict)


@object_exists(_get, class_plural)
def _delete(connection: LdapObject, distinguished_name: str) -> None:
    return _ldaps._delete(
        connection,
        # Un-Escape backslashes in distinguished names when modifying objects.
        ldap_escape(distinguished_name.replace("\\\\", "\\"), is_distinguished_name=True),
    )


@object_exists(_get, class_plural)
def _move(connection: LdapObject, distinguished_name: str, new_parent: str) -> None:
    return move(connection, distinguished_name, new_parent)


@object_exists(_get, class_plural)
def _modify(
    connection: LdapObject, distinguished_name: str, key: str, value: Union[int, str, bytes], operation: str
) -> None:
    return modify(
        connection,
        # Un-Escape backslashes in distinguished names when modifying objects.
        ldap_escape(distinguished_name.replace("\\\\", "\\"), is_distinguished_name=True),
        key,
        value,
        operation,
    )
