import copy
import logging
import functools

from typing import Callable, Union, Optional, List, Dict

from ldap.ldapobject import SimpleLDAPObject as LdapObject

from . import _ldaps, codecs, exceptions

UNSUPPORTED_SEARCH_ATTRIBUTES = ["thumbnailPhoto", "logonHours"]


logger = logging.getLogger(__name__)


class EscapedValue(str):
    pass


def ldap_escape(value: str, is_distinguished_name: Optional[bool] = False) -> EscapedValue:
    # Certain characters need to be replaced with their hexadecimal representation
    #
    # https://stackoverflow.com/a/39805523
    #
    # However if this is a distinguishedName, then there is an edge case where backslashes used to escape
    # trailing whitespace in a CN must *not be escaped*. I don't know why, welcome to Active Directory. See
    # the longer comment in format_distinguished_name for more information.
    if isinstance(value, EscapedValue):
        return value

    def _format_hex(char):
        return format(ord(char), "x")

    # This order must be retained to prevent recursive escaping of backslashes!
    ascii_map = {char: _format_hex(char) for char in ["\\", "(", ")", "*"]}

    new_characters = []
    for index, character in enumerate(value):
        new_character = copy.copy(character)
        replacement = ascii_map.get(character)
        if replacement is not None:
            # There is a replace for the current character...
            if character == "\\" and is_distinguished_name:
                # This is a backslash in a distinguishedName... check whether the proceeding
                # characters are " ,". If they are, the backslash is used to escape trailing
                # whitespace and *should not* be replaced with a hex code.
                if value[index : (index + 3)] == "\\ ,":
                    # This is a backslash used to escape whitespace, leave it...
                    logger.info(
                        f"not encoding backslash at position {index} in distinguishedName {value} "
                        "as it is there to escape a trailing space..."
                    )
                    new_character = character
                else:
                    # ...this one is used to escape whitespace, ldap-escape it.
                    new_character = f"\\{replacement}"
            else:
                new_character = f"\\{replacement}"
        new_characters.append(new_character)

    return EscapedValue("".join(new_characters))


def format_distinguished_name(value: str) -> str:
    # We found a user account at a customer who was created via the 'net user' command.
    #
    # Their username was in this format:
    #
    #   "firstname lastname "
    #
    # The trailing space gets *escaped with a backslash* by the 'net user' command, & when
    # PPA audits the user it sees this at the start of the DN:
    #
    #   "CN=firstname lastname\ "
    #
    # All operations performed against this user failed, & if it was a member of a group
    # PPA couldn't resolve it because the distinguishedName was being formatted badly.
    #
    # Because this isn't a 'real' backslash, it should not be escaped with another before
    # LDAP-escaping the query. So how do we know whether the backslash was supplied as part
    # of the username or not? If it was, then it should be escaped! Well I haven't found
    # any other way of creating a user with a trailing space other than 'net user', every
    # other tool strips leading & trailing whitespace. On top of that, 'net user' doesn't
    # let you supply a backslash at all, even if you escape it with another.
    #
    # With this in mind, if a user's DN starts with this...
    #
    #   "CN=firstname lastname\ "
    #
    # ...it's reasonably safe to say that the backslash *should not* be escaped with
    # another, but that all others in the DN should be escaped as-per the norm. The really
    # annoying thing is this can't be covered by our integration tests, because LDAP doesn't
    # let you create users in this format! However; we can at least cover it with unit tests.

    if "\\\\" in value:  # \\\\ would imply this value has already been escaped...
        return value

    new_value = []
    for index, character in enumerate(value):
        new_character = copy.copy(character)
        if character == "\\":
            # If this backslash *is not* followed by "  ," then it's a real one & should be escaped.
            if value[index : (index + 3)] != "\\ ,":
                new_character = "\\\\"
            else:
                logger.info(
                    f"not escaping backslash at position {index} in distinguishedName {value} "
                    "as it is there to escape a trailing space..."
                )
        new_value.append(new_character)
    return "".join(new_value)


def _encode_attributes(attr_dict: dict) -> Dict[str, Union[List, bytes]]:
    @functools.singledispatch
    def _encode(value):
        raise exceptions.ParameterError(
            f"Received LDAP attribute value {value} of unsupported type "
            f"{type(value)}. Values must be one of [int, str, bytes, list]."
        )

    @_encode.register
    def _int(value: int):
        return _encode(str(value))

    @_encode.register
    def _str(value: str):
        return bytes(value, "utf-8")

    @_encode.register
    def _bytes(value: bytes):
        return value

    @_encode.register
    def _list(value: list):
        return [_encode(v) for v in value]

    @_encode.register
    def _none(_: None):
        raise exceptions.ParameterError("None values cannot be supplied when creating an object.")

    return {key: _encode(value) for key, value in attr_dict.items()}


def validate_query_params(query_params):
    unsupported_attributes = ", ".join([key for key in query_params.keys() if key in UNSUPPORTED_SEARCH_ATTRIBUTES])
    if unsupported_attributes:
        raise exceptions.ParameterError(f"Search contains unsupported attributes {unsupported_attributes}.")


def _build_query_params(query_params) -> str:
    # Takes a full dictionary of attribute -> value pairs and returns an LDAP filter.

    @functools.singledispatch
    def _escape(value, key):
        raise exceptions.ParameterError(
            "Only string, integer, and None values are supported in search functions. "
            f"Received search attribute {key} with value {value} which is type {type(value)}."
        )

    @_escape.register
    def _str(value: str, key):

        escaper = functools.partial(ldap_escape, is_distinguished_name=False)

        if set(value) == {"*"}:
            raise exceptions.ParameterError(
                f"Query attribute {key} has value {value}, which is not supported for performance reasons."
            )

        wildcard_start, wildcard_end = list(map(lambda test: getattr(value, test)("*"), ["startswith", "endswith"]))

        # No wildcard at the start or the end - nothing more to do.
        if not any([wildcard_start, wildcard_end]):
            return f"({key}={escaper(value)})"
        else:
            if all([wildcard_start, wildcard_end]):
                formatted_value = f"({key}=*{escaper(value[1:-1])}*)"
            elif wildcard_start:
                formatted_value = f"({key}=*{escaper(value[1:])})"
            else:  # It must be at the end.
                formatted_value = f"({key}={escaper(value[:-1])}*)"
            return formatted_value

    @_escape.register
    def _none(_: None, key):
        return f"(!({key}=*))"

    @_escape.register
    def _int(value: int, key):
        return f"({key}={value})"

    query_chunks = []
    for key, value in query_params.items():

        if key == "distinguishedName":
            # When *searching* using a distinguishedName, any backslash must be
            # escaped with another backslash before being properly LDAP-escaped.
            value = format_distinguished_name(value)

        if any([isinstance(value, codecs.EncodedLdapBytes), isinstance(value, codecs.DecodedLdapString)]):
            raise exceptions.ParameterError(
                f"Attribute {key} has value {repr(value)}, which is not supported in LDAP queries. "
                "Please use a query encoder function to retrieve an EncodedQueryString object "
                "if this attribute requires a codec, or use a regular string as the value instead."
            )
        if any([isinstance(value, codecs.ProtectedString), isinstance(value, codecs.EncodedQueryString)]):
            # Neither a ProtectedValue or an EncodedQueryString should be escaped. These will either
            # have been generated by a query_encoder function in codecs.py, or created by the task author
            # for special attributes like objectGUID or objectSid.
            query_chunks.append(f"({key}={value})")
        else:
            query_chunks.append(_escape(value, key))

    return "".join(query_chunks)


def get(
    query_base: str,
    connection: LdapObject,
    search_base: Optional[str] = None,
    search_scope: Optional[int] = None,
    **query_params,
) -> list:
    search_filter = _build_query_params(query_params)
    search_query = f"(&{query_base}{search_filter})" if search_filter else query_base
    return _ldaps._search(connection, search_query, search_base=search_base, search_scope=search_scope)  # type: ignore


def object_exists(_get: Callable, class_plural: str = "objects"):
    def _decorator(func):
        """Decorates any function that assumes the target object
        already exists.

        This decorator also restricts functions so they can only
        operate against objects with the appropriate objectClass
        attribute. For example, an exception will be raised if you
        call users.delete with the distinguishedName of a group.

        The decorated function must take distinguished_name as the
        second positional argument.
        """

        @functools.wraps(func)
        def _fn(connection, distinguished_name: Union[List[str], str], *args, **kwargs):
            def _check_object_existence(dn):
                try:
                    _ = _get(connection, distinguishedName=dn)[0]
                except IndexError:
                    raise exceptions.NoSuchObject(f"No {class_plural} found with distinguishedName '{dn}'.")

            # Some functions (like users.not_member_of) will reference multiple
            # objects in a list, so we need to check the existence of all.
            if isinstance(distinguished_name, list):
                list(map(_check_object_existence, distinguished_name))
            else:
                _check_object_existence(distinguished_name)

            return func(connection, distinguished_name, *args, **kwargs)

        return _fn

    return _decorator


def modify(
    connection: LdapObject, distinguished_name: str, key: str, value: Union[int, str, bytes], operation: str
) -> None:
    ldap_operations = {"add": _ldaps._modify_add, "replace": _ldaps._modify_replace, "delete": _ldaps._modify_delete}

    if isinstance(value, int):
        value = str(value)
    if isinstance(value, str):
        value = bytes(value, "utf-8")

    try:
        ldap_operation = ldap_operations[operation]
    except KeyError:
        raise exceptions.ParameterError(
            f"Supplied operation {operation} not in supported list {list(ldap_operations.keys())}"
        )
    return ldap_operation(connection, distinguished_name, key, value)


def move(connection: LdapObject, distinguished_name: str, new_parent: str) -> None:
    distinguished_name = ldap_escape(distinguished_name.replace("\\\\", "\\"), is_distinguished_name=True)
    new_parent = new_parent.replace("\\\\", "\\")
    logger.debug(f"moving {distinguished_name} to {new_parent}")
    return _ldaps._move(connection, distinguished_name, new_parent)
