import os
import json
import logging
import functools

from typing import Dict, Optional, Any, List, Tuple

import requests

from . import exceptions

logger = logging.getLogger(__name__)

RESPONSE_EXCEPTIONS = {
    400: exceptions.BadRequest,
    500: exceptions.PPAServerError,
}


def event(name: str, text: Optional[str], files: Optional[List[Tuple]] = None, **kwargs):
    if files:
        return requests.post(f"{os.environ['PPA_URL']}/v1/{name}", dict(text=text, **kwargs), files=files)
    return requests.post(f"{os.environ['PPA_URL']}/v1/{name}", json=dict(text=text, **kwargs))


def task_event(status_code_exceptions: Optional[Dict[int, Any]] = None, event_type=None, return_tuple=None):
    def _raise_custom_exception(status_code: int, response_text: str):
        custom_exception = status_code_exceptions.get(status_code)
        if custom_exception is not None:
            raise custom_exception(event_type, response_text)

    def _wrapped_task_event(func):
        @functools.wraps(func)
        def _process_task_event(*args, **kwargs) -> Any:
            def _send_request(request_func) -> requests.Response:
                response = None
                try:
                    response = request_func()
                    response.raise_for_status()
                    return response
                except requests.exceptions.HTTPError as e:
                    # If the function has been decorated with custom exceptions then we may be able to match one
                    # of the status codes here. Note: not every non-200 status code will raise a urllib exception.
                    if status_code_exceptions is not None:
                        _raise_custom_exception(e.response.status_code, response.text)
                    try:
                        # Nothing provided in the decorator has caught the issue yet, so raise
                        # generic 400 and 500 exceptions if appropriate.
                        raise RESPONSE_EXCEPTIONS[e.response.status_code](str(e))
                    except KeyError:
                        # The request exception isn't handled and will be re-raised below.
                        pass
                    raise e

            response = _send_request(functools.partial(func, *args, **kwargs))
            if status_code_exceptions is not None:
                _raise_custom_exception(response.status_code, response.text)
            try:
                raise RESPONSE_EXCEPTIONS[response.status_code](response.text)
            except KeyError:
                pass
            # Any response that made it this far has been deemed acceptable. We don't care
            # about the status code anymore, just return response_json (which can be None
            # if the event is a UI output).
            try:
                response_json = response.json()
            except json.decoder.JSONDecodeError:
                return None

            return return_tuple(**response_json) if return_tuple else response_json

        return _process_task_event

    return _wrapped_task_event
